@@ -2,9 +2,12 @@
 
 import logging
 import re
+from hashlib import md5
+
+import six
 
 import cherrypy
-from cherrypy._cpcompat import basestring, md5, set, unicodestr
+from cherrypy._cpcompat import text_or_bytes
 from cherrypy.lib import httputil as _httputil
 from cherrypy.lib import is_iterator
 
@@ -30,7 +33,7 @@ def validate_etags(autotags=False, debug=False):
     response = cherrypy.serving.response
 
     # Guard against being run twice.
-    if hasattr(response, "ETag"):
+    if hasattr(response, 'ETag'):
         return
 
     status, reason, msg = _httputil.valid_status(response.status)
@@ -69,24 +72,24 @@ def validate_etags(autotags=False, debug=False):
         if debug:
             cherrypy.log('If-Match conditions: %s' % repr(conditions),
                          'TOOLS.ETAGS')
-        if conditions and not (conditions == ["*"] or etag in conditions):
-            raise cherrypy.HTTPError(412, "If-Match failed: ETag %r did "
-                                     "not match %r" % (etag, conditions))
+        if conditions and not (conditions == ['*'] or etag in conditions):
+            raise cherrypy.HTTPError(412, 'If-Match failed: ETag %r did '
+                                     'not match %r' % (etag, conditions))
 
         conditions = request.headers.elements('If-None-Match') or []
         conditions = [str(x) for x in conditions]
         if debug:
             cherrypy.log('If-None-Match conditions: %s' % repr(conditions),
                          'TOOLS.ETAGS')
-        if conditions == ["*"] or etag in conditions:
+        if conditions == ['*'] or etag in conditions:
             if debug:
                 cherrypy.log('request.method: %s' %
                              request.method, 'TOOLS.ETAGS')
-            if request.method in ("GET", "HEAD"):
+            if request.method in ('GET', 'HEAD'):
                 raise cherrypy.HTTPRedirect([], 304)
             else:
-                raise cherrypy.HTTPError(412, "If-None-Match failed: ETag %r "
-                                         "matched %r" % (etag, conditions))
+                raise cherrypy.HTTPError(412, 'If-None-Match failed: ETag %r '
+                                         'matched %r' % (etag, conditions))
 
 
 def validate_since():
@@ -110,7 +113,7 @@ def validate_since():
         since = request.headers.get('If-Modified-Since')
         if since and since == lastmod:
             if (status >= 200 and status <= 299) or status == 304:
-                if request.method in ("GET", "HEAD"):
+                if request.method in ('GET', 'HEAD'):
                     raise cherrypy.HTTPRedirect([], 304)
                 else:
                     raise cherrypy.HTTPError(412)
@@ -183,7 +186,7 @@ def proxy(base=None, local='X-Forwarded-Host', remote='X-Forwarded-For',
             # This is for lighttpd/pound/Mongrel's 'X-Forwarded-Proto: https'
             scheme = s
     if not scheme:
-        scheme = request.base[:request.base.find("://")]
+        scheme = request.base[:request.base.find('://')]
 
     if local:
         lbase = request.headers.get(local, None)
@@ -192,15 +195,14 @@ def proxy(base=None, local='X-Forwarded-Host', remote='X-Forwarded-For',
         if lbase is not None:
             base = lbase.split(',')[0]
     if not base:
+        base = request.headers.get('Host', '127.0.0.1')
         port = request.local.port
-        if port == 80:
-            base = '127.0.0.1'
-        else:
-            base = '127.0.0.1:%s' % port
+        if port != 80:
+            base += ':%s' % port
 
-    if base.find("://") == -1:
+    if base.find('://') == -1:
         # add http:// or https:// if needed
-        base = scheme + "://" + base
+        base = scheme + '://' + base
 
     request.base = base
 
@@ -210,7 +212,7 @@ def proxy(base=None, local='X-Forwarded-Host', remote='X-Forwarded-For',
             cherrypy.log('Testing remote %r:%r' % (remote, xff), 'TOOLS.PROXY')
         if xff:
             if remote == 'X-Forwarded-For':
-                #Bug #1268
+                # Bug #1268
                 xff = xff.split(',')[0].strip()
             request.remote.ip = xff
 
@@ -283,7 +285,7 @@ class SessionAuth(object):
 
     """Assert that the user is logged in."""
 
-    session_key = "username"
+    session_key = 'username'
     debug = False
 
     def check_username_and_password(self, username, password):
@@ -304,7 +306,7 @@ class SessionAuth(object):
 
     def login_screen(self, from_page='..', username='', error_msg='',
                      **kwargs):
-        return (unicodestr("""<html><body>
+        return (six.text_type("""<html><body>
 Message: %(error_msg)s
 <form method="post" action="do_login">
     Login: <input type="text" name="username" value="%(username)s" size="10" />
@@ -315,7 +317,7 @@ Message: %(error_msg)s
     <br />
     <input type="submit" />
 </form>
-</body></html>""") % vars()).encode("utf-8")
+</body></html>""") % vars()).encode('utf-8')
 
     def do_login(self, username, password, from_page='..', **kwargs):
         """Login. May raise redirect, or return True if request handled."""
@@ -324,15 +326,15 @@ Message: %(error_msg)s
         if error_msg:
             body = self.login_screen(from_page, username, error_msg)
             response.body = body
-            if "Content-Length" in response.headers:
+            if 'Content-Length' in response.headers:
                 # Delete Content-Length header so finalize() recalcs it.
-                del response.headers["Content-Length"]
+                del response.headers['Content-Length']
             return True
         else:
             cherrypy.serving.request.login = username
             cherrypy.session[self.session_key] = username
             self.on_login(username)
-            raise cherrypy.HTTPRedirect(from_page or "/")
+            raise cherrypy.HTTPRedirect(from_page or '/')
 
     def do_logout(self, from_page='..', **kwargs):
         """Logout. May raise redirect, or return True if request handled."""
@@ -354,56 +356,51 @@ Message: %(error_msg)s
         username = sess.get(self.session_key)
         if not username:
             sess[self.session_key] = username = self.anonymous()
-            if self.debug:
-                cherrypy.log(
-                    'No session[username], trying anonymous', 'TOOLS.SESSAUTH')
+            self._debug_message('No session[username], trying anonymous')
         if not username:
             url = cherrypy.url(qs=request.query_string)
-            if self.debug:
-                cherrypy.log('No username, routing to login_screen with '
-                             'from_page %r' % url, 'TOOLS.SESSAUTH')
+            self._debug_message(
+                'No username, routing to login_screen with from_page %(url)r',
+                locals(),
+            )
             response.body = self.login_screen(url)
-            if "Content-Length" in response.headers:
+            if 'Content-Length' in response.headers:
                 # Delete Content-Length header so finalize() recalcs it.
-                del response.headers["Content-Length"]
+                del response.headers['Content-Length']
             return True
-        if self.debug:
-            cherrypy.log('Setting request.login to %r' %
-                         username, 'TOOLS.SESSAUTH')
+        self._debug_message('Setting request.login to %(username)r', locals())
         request.login = username
         self.on_check(username)
 
+    def _debug_message(self, template, context={}):
+        if not self.debug:
+            return
+        cherrypy.log(template % context, 'TOOLS.SESSAUTH')
+
     def run(self):
         request = cherrypy.serving.request
         response = cherrypy.serving.response
 
         path = request.path_info
         if path.endswith('login_screen'):
-            if self.debug:
-                cherrypy.log('routing %r to login_screen' %
-                             path, 'TOOLS.SESSAUTH')
-            return self.login_screen(**request.params)
+            self._debug_message('routing %(path)r to login_screen', locals())
+            response.body = self.login_screen()
+            return True
         elif path.endswith('do_login'):
             if request.method != 'POST':
-                response.headers['Allow'] = "POST"
-                if self.debug:
-                    cherrypy.log('do_login requires POST', 'TOOLS.SESSAUTH')
+                response.headers['Allow'] = 'POST'
+                self._debug_message('do_login requires POST')
                 raise cherrypy.HTTPError(405)
-            if self.debug:
-                cherrypy.log('routing %r to do_login' % path, 'TOOLS.SESSAUTH')
+            self._debug_message('routing %(path)r to do_login', locals())
             return self.do_login(**request.params)
         elif path.endswith('do_logout'):
             if request.method != 'POST':
-                response.headers['Allow'] = "POST"
+                response.headers['Allow'] = 'POST'
                 raise cherrypy.HTTPError(405)
-            if self.debug:
-                cherrypy.log('routing %r to do_logout' %
-                             path, 'TOOLS.SESSAUTH')
+            self._debug_message('routing %(path)r to do_logout', locals())
             return self.do_logout(**request.params)
         else:
-            if self.debug:
-                cherrypy.log('No special path, running do_check',
-                             'TOOLS.SESSAUTH')
+            self._debug_message('No special path, running do_check')
             return self.do_check()
 
 
@@ -417,19 +414,19 @@ session_auth.__doc__ = """Session authentication hook.
 Any attribute of the SessionAuth class may be overridden via a keyword arg
 to this function:
 
-""" + "\n".join(["%s: %s" % (k, type(getattr(SessionAuth, k)).__name__)
-                 for k in dir(SessionAuth) if not k.startswith("__")])
+""" + '\n'.join(['%s: %s' % (k, type(getattr(SessionAuth, k)).__name__)
+                 for k in dir(SessionAuth) if not k.startswith('__')])
 
 
 def log_traceback(severity=logging.ERROR, debug=False):
     """Write the last error's traceback to the cherrypy error log."""
-    cherrypy.log("", "HTTP", severity=severity, traceback=True)
+    cherrypy.log('', 'HTTP', severity=severity, traceback=True)
 
 
 def log_request_headers(debug=False):
     """Write request headers to the cherrypy error log."""
-    h = ["  %s: %s" % (k, v) for k, v in cherrypy.serving.request.header_list]
-    cherrypy.log('\nRequest Headers:\n' + '\n'.join(h), "HTTP")
+    h = ['  %s: %s' % (k, v) for k, v in cherrypy.serving.request.header_list]
+    cherrypy.log('\nRequest Headers:\n' + '\n'.join(h), 'HTTP')
 
 
 def log_hooks(debug=False):
@@ -445,13 +442,13 @@ def log_hooks(debug=False):
             points.append(k)
 
     for k in points:
-        msg.append("    %s:" % k)
+        msg.append('    %s:' % k)
         v = request.hooks.get(k, [])
         v.sort()
         for h in v:
-            msg.append("        %r" % h)
+            msg.append('        %r' % h)
     cherrypy.log('\nRequest Hooks for ' + cherrypy.url() +
-                 ':\n' + '\n'.join(msg), "HTTP")
+                 ':\n' + '\n'.join(msg), 'HTTP')
 
 
 def redirect(url='', internal=True, debug=False):
@@ -536,7 +533,7 @@ def accept(media=None, debug=False):
     """
     if not media:
         return
-    if isinstance(media, basestring):
+    if isinstance(media, text_or_bytes):
         media = [media]
     request = cherrypy.serving.request
 
@@ -552,12 +549,12 @@ def accept(media=None, debug=False):
         # Note that 'ranges' is sorted in order of preference
         for element in ranges:
             if element.qvalue > 0:
-                if element.value == "*/*":
+                if element.value == '*/*':
                     # Matches any type or subtype
                     if debug:
                         cherrypy.log('Match due to */*', 'TOOLS.ACCEPT')
                     return media[0]
-                elif element.value.endswith("/*"):
+                elif element.value.endswith('/*'):
                     # Matches any subtype
                     mtype = element.value[:-1]  # Keep the slash
                     for m in media:
@@ -577,11 +574,11 @@ def accept(media=None, debug=False):
     # No suitable media-range found.
     ah = request.headers.get('Accept')
     if ah is None:
-        msg = "Your client did not send an Accept header."
+        msg = 'Your client did not send an Accept header.'
     else:
-        msg = "Your client sent this Accept header: %s." % ah
-    msg += (" But this resource only emits these media types: %s." %
-            ", ".join(media))
+        msg = 'Your client sent this Accept header: %s.' % ah
+    msg += (' But this resource only emits these media types: %s.' %
+            ', '.join(media))
     raise cherrypy.HTTPError(406, msg)
 
 
@@ -633,3 +630,19 @@ def autovary(ignore=None, debug=False):
         v.sort()
         resp_h['Vary'] = ', '.join(v)
     request.hooks.attach('before_finalize', set_response_header, 95)
+
+
+def convert_params(exception=ValueError, error=400):
+    """Convert request params based on function annotations, with error handling.
+
+    exception
+        Exception class to catch.
+
+    status
+        The HTTP error code to return to the client on failure.
+    """
+    request = cherrypy.serving.request
+    types = request.handler.callable.__annotations__
+    with cherrypy.HTTPError.handle(exception, error):
+        for key in set(types).intersection(request.params):
+            request.params[key] = types[key](request.params[key])
